from scurrypy import (
    Addon,
    Client,
    EmojiModel
)

class BotEmojisCacheAddon(Addon):
    """Defines caching bot emojis and lookup."""

    def __init__(self, client: Client):
        self.bot = client
        self.emojis: dict[str, EmojiModel] = {}   # index by unique name

        client.add_startup_hook(self.load_bot_emojis)

    async def load_bot_emojis(self):
        """Fetch all bot's emojis and add them to the cache."""
        emojis = await self.bot.bot_emoji().fetch_all()

        for emoji in emojis:
            self.emojis[emoji.name] = emoji

    def get_emoji(self, name: str):
        """Get an emoji from the cache.

        Args:
            name (str): name of the emoji

        Returns:
            (EmojiModel | None): the emoji object if found else None
        """
        return self.emojis.get(name)



from scurrypy import (
    Addon,
    Client
)

import inspect

class HooksAddon(Addon):
    """Defines registering startup and shutdown hooks with decorators."""

    def __init__(self, client: Client):
        self.bot = client

        self._start_hooks = []
        self._end_hooks = []

        client.add_startup_hook(self._run_start_hooks)
        client.add_shutdown_hook(self._run_end_hooks)

    def on_start(self, func):
        """Decorator to register startup hooks with params (bot)."""
        params_len = len(inspect.signature(func).parameters)
        if params_len != 1:
            raise TypeError(f"Startup hook handler '{func.__name__}' must accept exactly one parameter (bot).")

        self._start_hooks.append(func)
        return func
    
    def on_shutdown(self, func):
        params_len = len(inspect.signature(func).parameters)
        if params_len != 1:
            raise TypeError(f"Shutdown hook handler '{func.__name__}' must accept exactly one parameter (bot).")
        
        self._end_hooks.append(func)
        return func
    
    async def _run_start_hooks(self):
        for hook in self._start_hooks:
            result = hook(self.bot)
            if inspect.isawaitable(result):
                await result

    async def _run_end_hooks(self):
        for hook in self._end_hooks:
            result = hook(self.bot)
            if inspect.isawaitable(result):
                await result



from scurrypy import (
    Addon,
    Client,
    Intents,
    MessageCreateEvent,
    DiscordError
)

import inspect

class PrefixAddon(Addon):
    """Addon that implements automatic registering and decorating prefix commands."""

    def __init__(self, client: Client, prefix: str):
        """
        Args:
            client (Client): the Client object
            prefix (str): message prefix for commands
        """
        if not Intents.has(client.intents, Intents.MESSAGE_CONTENT):
            raise ValueError("Missing Intent.MESSAGE_CONTENT for scanning messages.")
        
        self.bot = client

        self.logger = client.logger

        self._prefix = prefix

        self._commands = {}
        """Maps prefix command names to handler."""
        
        client.add_event_listener("MESSAGE_CREATE", self.dispatch)

    def listen(self, name: str):
        """Listen for a prefix command.

        Args:
            name (str): name of the command
                !!! warning "Important"
                    Prefix commands are CASE-INSENSITIVE.
        """
        def decorator(func):
            params_len = len(inspect.signature(func).parameters)

            if params_len != 2:
                raise TypeError(
                    f"Prefix handler '{func.__name__}' must accept exactly two parameters (bot, message)."
                )
            self._commands[name.lower()] = func
        return decorator

    async def dispatch(self, event: MessageCreateEvent):
        """Dispatch event to user-defined handler.
            Ignore bot responding to self and messages without the desired prefix.

        Args:
            event (MessageCreateEvent): message create event object
        """
        if not event.content:
            return
        
        # ignore bot responding to itself
        if event.author.id == self.bot.application_id:
            return
        
        has_prefix = event.content.lower().startswith(self._prefix.lower())

        # ignore messages without prefix
        if not has_prefix:
            return
        
        command, *args = event.content[len(self._prefix):].strip().lower().split()
        handler = self._commands.get(command)

        # warn if this command doesnt have a known handler
        if not handler:
            self.logger.log_warn(f"Prefix Event '{command}' not found.")
            return

        # now prefix info can be confidently set
        try:
            res = self.bot.message(event.channel_id, event.id, context=event)
            await handler(self.bot, res)
            
            self.logger.log_info(f"Prefix Event '{command}' acknowledged with args: {list(args) or 'No args'}")
        except DiscordError as e:
            self.logger.log_error(f"Error in prefix command '{command}': {e}")        
        except Exception as e:
            self.logger.log_error(f"Unhandled error in prefix command '{command}': {e}")




from scurrypy import (
    Addon,
    Client,
    DiscordError,
    SlashCommand, UserCommand, MessageCommand, CommandOption,
    InteractionEvent, InteractionTypes, InteractionDataTypes
)

import inspect

def _check_func_params(func: callable):
    params_len = len(inspect.signature(func).parameters)

    if params_len != 2:
        raise TypeError(
            f"Interaction handler '{func.__name__}' must accept exactly two parameters (bot, interaction)."
        )

class SlashCommandAddon(Addon):
    """Addon that implements automatic registering and decorating interactions."""

    def __init__(self, client: Client, sync_commands: bool = True):
        """
        Args:
            client (Client): the bot client object
            sync_commands (bool): whether to sync commands. Defaults to `True`.
        """
        self.bot = client

        self.logger = client.logger

        self.sync_commands = sync_commands

        self._global_commands = []
        """List of all Global commands."""

        self._guild_commands = {}
        """Guild commands mapped by guild ID."""

        self.component_handlers = {}
        """Mapping of component custom IDs to handler."""

        self.slash_handlers = {}
        """Mapping of command names to handler."""

        self.message_handlers = {}
        """Mapping of message command names to handler."""

        self.user_handlers = {}
        """Mapping of user command names to handler."""

        self.autocomplete_handlers = {}
        """Mapping of autocomplete keys to handler."""

        client.add_startup_hook(self.on_startup) # wait until start to register commands

    def on_startup(self):
        """Sets up the addon with the client."""

        self.bot.add_event_listener('INTERACTION_CREATE', self.dispatch)
        if self.sync_commands:
            self.bot.add_startup_hook(self._register_commands)

    def component(self, custom_id: str):
        def decorator(func):
            _check_func_params(func)
            self.component_handlers[custom_id] = func
        return decorator
    
    # helpers purly for ergonomics
    def button(self, custom_id: str):
        """Decorator to route button interactions.

        Args:
            custom_id (str): custom ID of button
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
        """
        return self.component(custom_id)

    def select(self, custom_id: str):
        """Decorator to route select menu interactions.

        Args:
            custom_id (str): custom ID of select menu
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
        """
        return self.component(custom_id)

    def modal(self, custom_id: str):
        """Decorator to route modal interactions.

        Args:
            custom_id (str): custom ID of modal
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
        """
        return self.component(custom_id)

    def slash_command(self, 
        name: str, 
        description: str, 
        *, 
        options: list[CommandOption] = None, 
        guild_ids: list[int] = None
    ):
        """Register and route a slash command.

        Args:
            name (str): command name
            description (str): command description
            options (list[CommandOption], optional): list of command options
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        self._queue_command(SlashCommand(name, description, options), guild_ids)

        def decorator(func):
            _check_func_params(func)
            self.slash_handlers[name] = func
        return decorator
    
    def user_command(self, name: str, *, guild_ids: list[int] = None):
        """Register and route a user command.

        Args:
            name (str): command name
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        self._queue_command(UserCommand(name), guild_ids)

        def decorator(func):
            _check_func_params(func)
            self.user_handlers[name] = func
        return decorator

    def message_command(self, name: str, *, guild_ids: list[int] = None):
        """Register and route a message command.

        Args:
            name (str): command name
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        self._queue_command(MessageCommand(name), guild_ids)

        def decorator(func):
            _check_func_params(func)
            self.message_handlers[name] = func
        return decorator
    
    def autocomplete(self, command_name: str, option_name: str):
        """Register and route an autocomplete interaction.

        Args:
            command_name (str): name of command to autocomplete
            option_name (str): name of option to autocomplete
        """
        key = f"{command_name}:{option_name}"

        def decorator(func):
            _check_func_params(func)
            self.autocomplete_handlers[key] = func
        return decorator
    
    async def _register_commands(self):
        """Register both guild and global commands to the client."""

        if self._global_commands:
            await self.bot.register_global_commands(self._global_commands)

        for guild_id, cmds in self._guild_commands.items():
            await self.bot.register_guild_commands(cmds, guild_id)

        self.logger.log_info("Commands set!")
    
    def _queue_command(self, 
        command: SlashCommand | MessageCommand | UserCommand, 
        guild_ids: list[int] = None
    ):
        """Queue a decorated command to be registered on startup.

        Args:
            command (SlashCommand | MessageCommand | UserCommand): the command object
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        if guild_ids:
            gids = [guild_ids] if not isinstance(guild_ids, list) else guild_ids

            for gid in gids:
                self._guild_commands.setdefault(gid, []).append(command)
        
        else:
            self._global_commands.append(command)

    def clear_commands(self, guild_ids: list[int] = None):
        """Clear a guild's or global commands (slash, message, and user).

        Args:
            guild_ids (list[int], optional): list of guild IDs for guild commands or omit for global
        """
        if guild_ids:
            gids = [guild_ids] if isinstance(guild_ids, int) else guild_ids
            for gid in gids:
                removed = self._guild_commands.pop(gid, None)
                if removed is None:
                    self.logger.log_warn(f"Guild ID {gid} not found; skipping...")
        else:
            self._global_commands.clear()

    def _get_handler(self, name: str):
        """Helper function for fetching a handler by `fnmatch`."""

        import fnmatch
        for k, v in self.component_handlers.items():
            if fnmatch.fnmatch(name, k):
                return v
        return False

    async def dispatch(self, event: InteractionEvent):
        """Dispatch a response to an `INTERACTION_CREATE` event

        Args:
            event (InteractionEvent): interaction event object
        """
        handler = None
        name = None

        match event.type:
            case InteractionTypes.APPLICATION_COMMAND:
                name = event.data.name
                match event.data.type:
                    case InteractionDataTypes.SLASH_COMMAND:
                        handler = self.slash_handlers.get(name)
                    case InteractionDataTypes.USER_COMMAND:
                        handler = self.user_handlers.get(name)
                    case InteractionDataTypes.MESSAGE_COMMAND:
                        handler = self.message_handlers.get(name)

            # BOTH modals and message components have custom IDs!
            case InteractionTypes.MESSAGE_COMPONENT | InteractionTypes.MODAL_SUBMIT:
                name = event.data.custom_id
                handler = self._get_handler(name)

            case InteractionTypes.APPLICATION_COMMAND_AUTOCOMPLETE:
                # Extract option being autocompleted

                focused = next((opt for opt in event.data.options if opt.focused), None)

                if not focused:
                    self.logger.log_error("No focused option found for autocomplete!")
                    return

                name = f"{event.data.name}:{focused.name}"
                handler = self.autocomplete_handlers.get(name)

        if not handler:
            self.logger.log_warn(f"No handler registered for interaction '{name}'")
            return

        try:
            res = self.bot.interaction(event.id, event.token, context=event)
            await handler(self.bot, res)
            self.logger.log_info(f"Interaction '{name}' Acknowledged.")
        except DiscordError as e:
            self.logger.log_error(f"Error in interaction '{name}': {e}")
        except Exception as e:
            self.logger.log_error(f"Unhandled error in interaction '{name}': {e}")



from scurrypy import (
    Addon,
    Client,
    DiscordError,
    ChannelModel,
    GuildCreateEvent, GuildDeleteEvent,
    GuildChannelCreateEvent, GuildChannelUpdateEvent, GuildChannelDeleteEvent
)

class GuildChannelCacheAddon(Addon):
    """Defines caching channels and lookup."""

    def __init__(self, client: Client):
        self.bot = client

        self.channels: dict[int, dict[int, ChannelModel]] = {}  # stores OBJECTS
        self.channel_index: dict[int, ChannelModel] = {}        # stores REFERENCES

        client.add_event_listener('GUILD_CREATE', self.on_guild_create)
        client.add_event_listener('GUILD_DELETE', self.on_guild_delete)

        client.add_event_listener('CHANNEL_CREATE', self.on_channel_create)
        client.add_event_listener('CHANNEL_UPDATE', self.on_channel_update)
        client.add_event_listener('CHANNEL_DELETE', self.on_channel_delete)

    def on_guild_create(self, event: GuildCreateEvent):
        """Append new guild channels to cache. Also add channels to index.

        Args:
            event (GuildCreateEvent): the GUILD_CREATE event
        """
        guild_dict = self.channels.setdefault(event.id, {})

        for ch in event.channels:
            guild_dict[ch.id] = ch
            self.channel_index[ch.id] = ch

    def on_guild_delete(self, event: GuildDeleteEvent):
        """Remove guild channels from cache. Also remove channels from index

        Args:
            event (GuildDeleteEvent): the GUILD_DELETE event
        """
        removed_channels = self.channels.pop(event.id, {})

        for ch in removed_channels.values():
            self.channel_index.pop(ch.id, None)

    def on_channel_create(self, event: GuildChannelCreateEvent):
        """Append channel to guild key. Also append channel to index.

        Args:
            event (GuildChannelCreateEvent): the CHANNEL_CREATE event
        """
        model = ChannelModel.from_dict(event.raw)
        guild_dict = self.channels.setdefault(event.guild_id, {})

        guild_dict[event.id] = model
        self.channel_index[event.id] = model

    def on_channel_update(self, event: GuildChannelUpdateEvent):
        """Replace channel in guild key. Also replace channel in index.

        Args:
            event (GuildChannelUpdateEvent): the CHANNEL_UPDATE event
        """
        model = ChannelModel.from_dict(event.raw)
        guild_dict = self.channels.setdefault(event.guild_id, {})

        guild_dict[event.id] = model
        self.channel_index[event.id] = model

    def on_channel_delete(self, event: GuildChannelDeleteEvent):
        """Remove channel from guild key. Also remove channel from index.

        Args:
            event (GuildChannelDeleteEvent): the CHANNEL_DELETE event
        """
        model = self.channel_index.pop(event.id, None)
        if model:
            self.channels.get(event.guild_id, {}).pop(event.id, None)

    async def get_channel(self, channel_id: int):
        """Fetch a guild channel. If not found, request and store it.

        Args:
            channel_id (int): ID of channel

        Returns:
            (ChannelModel | None): hydrated channel object or None if fetch failed
        """
        channel = self.channel_index.get(channel_id)
        if channel:
            return channel

        try:
            channel = await self.bot.channel(channel_id).fetch()
        except DiscordError:
            return None

        self.put(channel)

        return channel

    def put(self, channel: ChannelModel):
        """Put a new channel into the cache.

        Args:
            channel (ChannelModel): the channel object

        Raises:
            ValueError: missing `guild_id`
        """
        if channel.guild_id is None:
            raise ValueError("Cannot cache a channel without a guild_id.")
        
        guild_dict = self.channels.setdefault(channel.guild_id, {})
        guild_dict[channel.id] = channel
        self.channel_index[channel.id] = channel



from scurrypy import (
    Addon,
    Client,
    Intents,
    EmojiModel,
    GuildCreateEvent, GuildDeleteEvent,
    GuildEmojisUpdateEvent
)

class GuildEmojiCacheAddon(Addon):
    """Defines caching guild emojis and lookup.
    
    !!! important
        This cache requires `Intents.GUILD_EMOJIS_AND_STICKERS` to keep up-to-date.
    """

    def __init__(self, client: Client):
        self.bot = client

        if not Intents.has(client.intents, Intents.GUILD_EMOJIS_AND_STICKERS):
            raise ValueError("GuildEmojiCache requires Intents.GUILD_EMOJIS_AND_STICKERS for GUILD_EMOJIS_UPDATE event.")

        self.guild_emojis: dict[int, dict[int, EmojiModel]] = {}  # owns emoji objects
        self.guild_emoji_index: dict[int, EmojiModel] = {}        # index by ID (reference)

        client.add_event_listener('GUILD_CREATE', self.on_guild_create)
        client.add_event_listener('GUILD_DELETE', self.on_guild_delete)

        client.add_event_listener('GUILD_EMOJIS_UPDATE', self.on_emojis_update)

    def on_guild_create(self, event: GuildCreateEvent):
        """Append new guild emojis to cache. Also add emojis to index.

        Args:
            event (GuildCreateEvent): the GUILD_CREATE event
        """
        guild_dict = self.guild_emojis.setdefault(event.id, {})

        # event.emojis is already hydrated
        for emoji in event.emojis:
            guild_dict[emoji.id] = emoji
            self.guild_emoji_index[emoji.id] = emoji

    def on_guild_delete(self, event: GuildDeleteEvent):
        """Remove guild emojis from cache. Also remove emojis from index

        Args:
            event (GuildDeleteEvent): the GUILD_DELETE event
        """
        removed = self.guild_emojis.pop(event.id, {})

        for emoji in removed.values():
            self.guild_emoji_index.pop(emoji.id, None)

    def on_emojis_update(self, event: GuildEmojisUpdateEvent):
        """Refresh guild emojis with new list. Also refresh the index

        Args:
            event (GuildEmojisUpdateEvent): the GUILD_EMOJIS_UPDATE event
        """
        guild_id = event.guild_id

        # remove old emojis
        removed = self.guild_emojis.pop(guild_id, {})

        for emoji in removed.values():
            self.guild_emoji_index.pop(emoji.id, None)

        # add new emoji set (full replacement)
        guild_dict = self.guild_emojis.setdefault(guild_id, {})

        for emoji in event.emojis:
            guild_dict[emoji.id] = emoji
            self.guild_emoji_index[emoji.id] = emoji

    def get_emoji(self, emoji_id: int):
        """Get an emoji from the cache.

        Args:
            emoji_id (int): ID of the emoji

        Returns:
            (EmojiModel | None): the Emoji object if found, else None
        """
        return self.guild_emoji_index.get(emoji_id, None)
    


# scurry_kit/builders

from .action_rows import ActionRowBuilder
from .embeds import EmbedBuilder
from .container import ContainerBuilder
from .modals import LabelBuilder

__all__ = [
    "ActionRowBuilder",
    "EmbedBuilder",
    "ContainerBuilder",
    "LabelBuilder"
]



from scurrypy import (
    Addon,
    UserModel,
    EmbedAuthor
)

class EmbedBuilder(Addon):
    """Common Embed helpers."""
    
    @staticmethod
    def user_author(user: UserModel):
        """Embed author builder.

        Args:
            user (UserModel): user author

        Returns:
            (EmbedAuthor): the EmbedAuthor object
        """
        if not user:
            raise ValueError("Missing user.")
        
        return EmbedAuthor(
            name=user.username,
            icon_url=f"https://cdn.discordapp.com/avatars/{user.id}/{user.avatar}.png"
        )
    
    @staticmethod
    def timestamp():
        """Embed timestamp builder. Adheres to ISO8601 format.

        Returns:
            (str): message timestamp
        """
        from datetime import datetime, timezone

        return datetime.now(timezone.utc).isoformat()



from scurrypy import (
    Addon,
    Client,
    DiscordError,
    RoleModel,
    GuildCreateEvent, GuildDeleteEvent,
    RoleCreateEvent, RoleUpdateEvent, RoleDeleteEvent
)

class RoleCacheAddon(Addon):
    """Defines caching guild roles and lookup."""

    def __init__(self, client: Client):
        self.bot = client

        self.roles: dict[int, dict[int, RoleModel]] = {} # stores OBJECTS
        self.role_index: dict[int, RoleModel] = {} # stores REFERENCES

        client.add_event_listener('GUILD_CREATE', self.on_guild_create)
        client.add_event_listener('GUILD_DELETE', self.on_guild_delete)

        client.add_event_listener('ROLE_CREATE', self.on_role_create)
        client.add_event_listener('ROLE_UPDATE', self.on_role_update)
        client.add_event_listener('ROLE_DELETE', self.on_role_delete)

    def on_guild_create(self, event: GuildCreateEvent):
        """Append new guild roles to cache. Also add roles to index.

        Args:
            event (GuildCreateEvent): the GUILD_CREATE event
        """
        guild_dict = self.roles.setdefault(event.id, {})

        for role in event.roles:
            guild_dict[role.id] = role
            self.role_index[role.id] = role

    def on_guild_delete(self, event: GuildDeleteEvent):
        """Remove guild roles from cache. Also remove roles from index

        Args:
            event (GuildDeleteEvent): the GUILD_DELETE event
        """
        removed_roles = self.roles.pop(event.id, {})

        for role in removed_roles.values():
            self.role_index.pop(role.id, None)

    def on_role_create(self, event: RoleCreateEvent):
        """Append role to guild key. Also append role to index.

        Args:
            event (RoleCreateEvent): the ROLE_CREATE event
        """
        model = RoleModel.from_dict(event.raw)
        guild_dict = self.roles.setdefault(event.guild_id, {})

        guild_dict[event.role.id] = model
        self.role_index[event.role.id] = model

    def on_role_update(self, event: RoleUpdateEvent):
        """Replace role in guild key. Also replace role in index.

        Args:
            event (RoleUpdateEvent): the ROLE_UPDATE event
        """
        model = RoleModel.from_dict(event.raw)
        guild_dict = self.roles.setdefault(event.guild_id, {})

        guild_dict[event.role.id] = model
        self.role_index[event.role.id] = model

    def on_role_delete(self, event: RoleDeleteEvent):
        """Remove role from guild key. Also remove role from index.

        Args:
            event (RoleDeleteEvent): the ROLE_DELETE event
        """
        model = self.role_index.pop(event.role_id, None)
        if model:
            self.roles.get(event.guild_id, {}).pop(event.role_id, None)

    def get_role(self, role_id: int):
        """Get a role from the cache.

        Args:
            role_id (int): ID of the role

        Returns:
            (RoleModel | None): the role object if found else None
        """
        return self.role_index.get(role_id)
    
    async def get_role(self, guild_id: int, role_id: int):
        """Fetch a guild role. If not found, request and store it.

        Args:
            guild_id (int): guild ID of role
            role_id (int): role ID of guild

        Returns:
            (RoleModel | None): hydrated role object or None if fetch failed
        """
        role = self.role_index.get(role_id)
        if role:
            return role
        
        try:
            role = await self.bot.guild(guild_id).fetch_guild_role(role_id)
        except DiscordError:
            return None
        
        self.put(guild_id, role)
        return role

    def put(self, guild_id: int, role: RoleModel):
        """Put a new role into the cache.

        Args:
            guild_id (int): guild ID of the role
            role (RoleModel): the role object
        """
        guild_dict = self.roles.setdefault(guild_id, {})
        guild_dict[role.id] = role
        self.role_index[role.id] = role



# scurry_kit/addons/cache

from .bot_emojis import BotEmojisCacheAddon
from .channel import GuildChannelCacheAddon
from .guild_emojis import GuildEmojiCacheAddon
from .role import RoleCacheAddon

__all__ = [
    'BotEmojisCacheAddon',
    'GuildChannelCacheAddon',
    'GuildEmojiCacheAddon',
    'RoleCacheAddon'
]



from scurrypy import (
    Addon,
    ActionRowPart, ActionRowChild,
    EmojiModel,
    Button, ButtonStyles
)

class ActionRowBuilder(Addon):
    """Common button helpers."""
    
    @staticmethod
    def _basic_button(
        style: int, 
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        if emoji:
            if isinstance(emoji, str):
                emoji = EmojiModel(name=emoji)
            elif not isinstance(emoji, EmojiModel):
                raise TypeError(f"EasyBot.primary expects type str or EmojiModel, got {type(emoji).__name__}")
        
        return Button(
            style=style,
            custom_id=custom_id,
            label=label,
            emoji=emoji ,
            disabled=disabled
        )
    
    @staticmethod
    def primary(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a primary button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ActionRowBuilder._basic_button(ButtonStyles.PRIMARY, custom_id, label, emoji, disabled)
    
    @staticmethod
    def secondary(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a secondary button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ActionRowBuilder._basic_button(ButtonStyles.SECONDARY, custom_id, label, emoji, disabled)
    
    @staticmethod
    def success(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a success button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ActionRowBuilder._basic_button(ButtonStyles.SUCCESS, custom_id, label, emoji, disabled)
    
    @staticmethod
    def danger(
        custom_id: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a danger button

        Args:
            custom_id (str): unique button identifier
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        return ActionRowBuilder._basic_button(ButtonStyles.DANGER, custom_id, label, emoji, disabled)
    
    @staticmethod
    def link(
        url: str, 
        label: str = None, 
        emoji: str | EmojiModel = None, 
        disabled: bool = False
    ):
        """Builds a link button

        Args:
            url (str): button URL to open
            label (str, optional): user-facing label
            emoji (str | EmojiModel, optional): emoji icon as str or EmojiModel if custom
            disabled (bool, optional): Whether the button should be disabled. Defaults to False.

        Returns:
            (Button): the button object
        """
        btn = ActionRowBuilder._basic_button(ButtonStyles.LINK, label, emoji, disabled)
        btn.url = url
        return btn

    @staticmethod
    def row(components: list[ActionRowChild]):
        """Builds an action row.

        Args:
            components (list[ActionRowChild]): components to be placed in the row

        Returns:
            (ActionRowPart): the action row object
        """
        if not isinstance(components, list):
            components = [components]
        
        return ActionRowPart(components)



# scurry_kit/addons

from .commands import SlashCommandAddon
from .events import EventsAddon
from .hooks import HooksAddon
from .prefix import PrefixAddon

__all__ = [
    "SlashCommandAddon",
    "EventsAddon",
    "HooksAddon",
    "PrefixAddon"
]



from scurrypy import (
    Addon,
    TextInput, TextInputStyles
)

class LabelBuilder(Addon):
    """Common Label Helpers."""
    
    @staticmethod
    def _text(
        custom_id: str,
        style: int,
        min_length: int,
        max_length: int,
        required: bool,
        value: str,
        placeholder: str
    ):
        return TextInput(
            custom_id=custom_id,
            style=style,
            min_length=min_length,
            max_length=max_length,
            required=required,
            value=value,
            placeholder=placeholder
        )
    
    @staticmethod
    def short_text(
        custom_id: str,
        min_length: int = None,
        max_length: int = None,
        required: bool = True,
        value: str = None,
        placeholder: str = None
    ):
        """Builds a single-line text input.

        Args:
            custom_id (str): ID for the input
            min_length (int): minimum input length
            max_length (int): maximum input length
            required (bool): Whether this input is required. Defaults to True.
            value (str): Pre-filled content
            placeholder (str): placeholder if text is empty

        Returns:
            (TextInput): the text input object
        """
        return LabelBuilder._text(custom_id, TextInputStyles.SHORT, min_length, max_length, required, value, placeholder)
    
    @staticmethod
    def long_text(
        custom_id: str,
        min_length: int,
        max_length: int,
        required: bool,
        placeholder: str
    ):
        """Builds a multi-line text input.

        Args:
            custom_id (str): ID for the input
            min_length (int): minimum input length
            max_length (int): maximum input length
            required (bool): Whether this input is required. Defaults to True.
            value (str): Pre-filled content
            placeholder (str): placeholder if text is empty

        Returns:
            (TextInput): the text input object
        """
        return LabelBuilder._text(custom_id, TextInputStyles.PARAGRAPH, min_length, max_length, required, placeholder)



from scurrypy import (
    Addon,
    Separator, SeparatorTypes
)

class ContainerBuilder(Addon):
    """Common Container Helpers."""
    
    @staticmethod
    def small_separator(divider: bool = True):
        """Builds a separator with a small padding.

        Args:
            divider (bool, optional): Whether a visual divider should appear. Defaults to True.

        Returns:
            (Separator): the separator object
        """
        return Separator(divider, SeparatorTypes.SMALL_PADDING)
    
    @staticmethod
    def large_separator(divider: bool = True):
        """Builds a separator with a large padding.

        Args:
            divider (bool, optional): Whether a visual divider should appear. Defaults to True.

        Returns:
            (Separator): the separator object
        """
        return Separator(divider, SeparatorTypes.LARGE_PADDING)



from scurrypy import (
    Addon,
    Client,
    DiscordError,
    Event
)

import inspect

class EventsAddon(Addon):
    """Addon that implements automatic registering and decorating events."""

    def __init__(self, client: Client):
        """
        Args:
            client (Client): the Client object
        """
        self.bot = client

        self.logger = client.logger

        self._events = {}
        """Maps EVENT_NAME to handlers."""

        client.add_startup_hook(self.on_startup)

    def on_startup(self):
        """Adds registered events to client's event listener."""

        # lead all registered events to this dispatch
        for dispatch_type in self._events.keys():
            self.bot.add_event_listener(dispatch_type, self.dispatch)
        
    def listen(self, event_name: str):
        """Listen for an event in which to listen.

        Args:
            event_name (str): event name
        """
        def decorator(func):
            params_len = len(inspect.signature(func).parameters)
            if params_len != 2:
                raise TypeError(f"Event handler '{func.__name__}' must accept exactly two parameters (bot, event).")
            
            self._events.setdefault(event_name, []).append(func)
        return decorator

    async def dispatch(self, event: Event):
        """Addon's entry point.

        Args:
            event (Event): event data object
        """
        handlers = self._events.get(event.name)

        if not handlers:
            return
        try:
            for handler in handlers:
                await handler(self.bot, event)
        except DiscordError as e:
            self.logger.log_error(f"Error in event '{handler}': {e}")
        except Exception as e:
            self.logger.log_error(f"Unhandled error in event '{handler.__name__}': {e}")
